#include "bitmap.h"
#include "jpeglib/jpeglib.h"
#include "libpng/png.h"

void CopyRGBtoBGRA(byte *dest, byte *source, int pixels)
{
	for (int i = 0; i < pixels; i++)
	{
		*dest++ = *(source + 2);
		*dest++ = *(source + 1);
		*dest++ = *(source + 0);
		*dest++ = 0;
		source += 3;
	}
}

void CopyGrayscaletoBGRA(byte *dest, byte *source, int pixels)
{
	for (int i = 0; i < pixels; i++)
	{
		*dest++ = *source;
		*dest++ = *source;
		*dest++ = *source;
		*dest++ = 0;
		source++;
	}
}

void CopyBGRAtoRGB(byte *dest, byte *source, int pixels)
{
	for (int i = 0; i < pixels; i++)
	{
		*dest++ = *(source + 2);
		*dest++ = *(source + 1);
		*dest++ = *(source + 0);
		source += 4;
	}
}

void Bitmap::CreateBitmap(int cx, int cy)
{
	this->cx = cx;
	this->cy = cy;

	ZeroMemory(&bmih, sizeof(BITMAPINFOHEADER));
	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biWidth = cx;
	bmih.biHeight = cy;
	bmih.biPlanes = 1;
	bmih.biBitCount = 32;
	bmih.biCompression = BI_RGB;

	hbmBitmap = CreateDIBSection(NULL, (LPBITMAPINFO)(&bmih), DIB_RGB_COLORS, (VOID**)&pBitmap, NULL, 0);
	hdcBitmap = CreateCompatibleDC(0);
	SelectObject(hdcBitmap, hbmBitmap);
	SelectObject(hdcBitmap, GetStockObject(DC_PEN));
	SelectObject(hdcBitmap, GetStockObject(DC_BRUSH));
	SetBkMode(hdcBitmap, TRANSPARENT);

	if (hbmBitmap == NULL || hdcBitmap == NULL)
		throw L"Error create bitmap! (CreateBitmap(int cx, int cy))";
}

void Bitmap::DeleteBitmap()
{
	DeleteDC(hdcBitmap);
	DeleteObject(hbmBitmap);
	hdcBitmap = NULL;
	hbmBitmap = NULL;
	pBitmap = NULL;
	cx = 0;
	cy = 0;
}

Bitmap::Bitmap()
{
	cx = 0;
	cy = 0;
	pBitmap = NULL;
	hbmBitmap = NULL;
	hdcBitmap = NULL;
}

Bitmap::Bitmap(int cx, int cy)
{
	CreateBitmap(cx, cy);
}

Bitmap::Bitmap(wchar *fileName)
{
	if (!LoadFromFile(fileName))
		throw L"Error open file";
}

Bitmap::~Bitmap()
{
	DeleteBitmap();
}


bool Bitmap::LoadFromFile(wchar *fileName)
{
	String fn = fileName;
	fn.ToUpperCase();

	if (fn.SubString(L".BMP"))
		return LoadFromBmp(fileName);
	else if (fn.SubString(L".JPEG") || fn.SubString(L".JPG"))
		return LoadFromJpeg(fileName);
	else if (fn.SubString(L".PNG"))
		return LoadFromPng(fileName);
	else
		throw L"Unknown bitmap file format";
}

bool Bitmap::SaveToFile(wchar *fileName)
{
	String fn = fileName;
	fn.ToUpperCase();

	if (fn.SubString(L".BMP"))
		return SaveToBmp(fileName);
	else if (fn.SubString(L".JPEG") || fn.SubString(L".JPG"))
		return SaveToJpeg(fileName);
	else if (fn.SubString(L".PNG"))
		return SaveToPng(fileName);
	else
		throw L"Unknown bitmap file format";
}

bool Bitmap::LoadFromBmp(wchar *fileName)
{
	HANDLE hfile = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL); 
	if (hfile == INVALID_HANDLE_VALUE)
		return false;
	DWORD nBytesRead;

	BITMAPFILEHEADER bmf;
	ReadFile(hfile, &bmf, sizeof(bmf), &nBytesRead, NULL);
	if (nBytesRead != sizeof(bmf))
		return false;

	ReadFile(hfile, &bmih, sizeof(bmih), &nBytesRead, NULL);
	if (nBytesRead != sizeof(bmih))
		return false;

	if (bmf.bfType != 0x4D42 || bmih.biSize != sizeof(bmih) || 
		bmih.biBitCount != 32 && bmih.biBitCount != 24 || 
		bmih.biPlanes != 1 || bmih.biCompression != BI_RGB)
		throw L"Must be 32 bit bitmap";

	WORD bitCount = bmih.biBitCount;

	CreateBitmap(bmih.biWidth, bmih.biHeight);

	if (bitCount == 32)
	{
		ReadFile(hfile, pBitmap, cx*cy*4, &nBytesRead, NULL);
		if (nBytesRead != cx*cy*4)
			return false;
	}
	else if (bitCount == 24)
	{
		int nLine = cx*3;
		while (nLine%4 != 0) 
			nLine++;
		
		BYTE *data = new BYTE[nLine];
		for (int j = 0; j < cy; j++)
		{
			ReadFile(hfile, data, nLine, &nBytesRead, NULL);
			if (nBytesRead != nLine)
				return false;

			PBYTE dataLine = data;
			PBYTE bitmapLine = pBitmap + 4*cx*j;
			for (int i = 0; i < cx; i++)
			{
				*bitmapLine++ = *dataLine++;
				*bitmapLine++ = *dataLine++;
				*bitmapLine++ = *dataLine++;
				*bitmapLine++ = 0;
			}
		}
		delete data;
	}
	return true;
}

bool Bitmap::SaveToBmp(wchar *fileName)
{
	HANDLE hfile = CreateFile(fileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL); 
	if (hfile == INVALID_HANDLE_VALUE)
		return false;
	DWORD nBytesWritten;
	
	BITMAPFILEHEADER bmf;
	bmf.bfType = 0x4D42;
	bmf.bfSize = 0;
	bmf.bfReserved1 = 0;
	bmf.bfReserved2 = 0;
	bmf.bfOffBits = sizeof(BITMAPFILEHEADER);
	WriteFile(hfile, &bmf, sizeof(bmf), &nBytesWritten, NULL);

	WriteFile(hfile, &bmih, sizeof(bmih), &nBytesWritten, NULL);

	WriteFile(hfile, pBitmap, cx*cy*4, &nBytesWritten, NULL);

	CloseHandle(hfile);
	return true;
}


bool Bitmap::LoadFromJpeg(wchar *fileName)
{
	FILE *infile;
	infile = _wfopen(fileName, L"rb");
	if (!infile)
		return false;

	jpeg_decompress_struct cinfo;
	jpeg_error_mgr jerr;
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, infile);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	CreateBitmap(cinfo.image_width, cinfo.image_height);

	byte *data = new byte[cx*3];
	while (cinfo.output_scanline < cinfo.image_height)
	{
		byte *bitmapLine = GetScanLine(cinfo.output_scanline);
		jpeg_read_scanlines(&cinfo, (JSAMPARRAY)&data, 1);
		if (cinfo.num_components == 3)
			CopyRGBtoBGRA(bitmapLine, data, cx);
		else
			CopyGrayscaletoBGRA(bitmapLine, data, cx);
	}
	delete data;

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);
	fclose(infile);
	return true;
}

bool Bitmap::SaveToJpeg(wchar *fileName)
{
	FILE *outfile;
	outfile = _wfopen(fileName, L"wb");
	if (!outfile)
		return false;

	jpeg_compress_struct cinfo;
	jpeg_error_mgr jerr;
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo, outfile);

	cinfo.image_width = cx;
	cinfo.image_height = cy;
	cinfo.input_components = 3;
	cinfo.in_color_space = JCS_RGB;

	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, iDefJpegQuality, TRUE);
	jpeg_start_compress(&cinfo, TRUE);

	byte *data = new byte[cx*3];
	while (cinfo.next_scanline < cinfo.image_height)
	{
		byte *bitmapLine = GetScanLine(cinfo.next_scanline);
		CopyBGRAtoRGB(data, bitmapLine, cx);
		jpeg_write_scanlines(&cinfo, (JSAMPARRAY)&data, 1);
	}
	delete data;

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);
	fclose(outfile);
	return true;
}


void png_user_error_fn(png_structp png_ptr, png_const_charp error_msg)
{
	String error = (char*)error_msg;
	throw error.GetString();
}

void png_user_warning_fn(png_structp png_ptr, png_const_charp warning_msg)
{
	String warning = (char*)warning_msg;
	MessageBox(NULL, L"PNG Warning!", warning.GetString(), MB_ICONWARNING);
}

#define PNG_BYTES_TO_CHECK 4
bool Bitmap::LoadFromPng(wchar *fileName)
{
	FILE *file = _wfopen(fileName, L"rb");
	if (!file)
		return false;

	png_structp png_ptr;

	char header[PNG_BYTES_TO_CHECK];
	fread(header, 1, PNG_BYTES_TO_CHECK, file);
	if (png_sig_cmp((png_bytep)header, (png_size_t)0, PNG_BYTES_TO_CHECK))
		return false;

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, 
		&png_user_error_fn, &png_user_warning_fn);
	png_infop info_ptr = png_create_info_struct(png_ptr);
	png_infop end_info_ptr = png_create_info_struct(png_ptr);
	png_init_io(png_ptr, file);
	png_set_sig_bytes(png_ptr, PNG_BYTES_TO_CHECK);
    
	// low-level read
	png_read_info(png_ptr, info_ptr);

	png_uint_32 width = png_get_image_width(png_ptr, info_ptr);
	png_uint_32 height = png_get_image_height(png_ptr, info_ptr);
	int bit_depth = png_get_bit_depth(png_ptr, info_ptr);
	int color_type = png_get_color_type(png_ptr, info_ptr);
	int interlace_type = png_get_interlace_type(png_ptr, info_ptr);
	if (bit_depth != 8 || interlace_type != PNG_INTERLACE_NONE || color_type != PNG_COLOR_TYPE_RGB)
		throw L"Error open PNG: Supported only none interlaced 8 bit RGB and RGBA images";

	CreateBitmap(width, height);

	png_byte *row = new png_byte[width*4];
	png_read_update_info(png_ptr, info_ptr);
	for (int j = 0; j < cy; j++)
	{
		png_read_row(png_ptr, row, NULL);
		CopyRGBtoBGRA(GetScanLine(j), row, width);
	}
	png_read_end(png_ptr, info_ptr);
	delete row;

	png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	fclose(file);
	return true;
}

bool Bitmap::SaveToPng(wchar *fileName)
{
	FILE *file = _wfopen(fileName, L"wb");
	if (!file)
		return false;

	png_structp png_ptr;
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, 
		&png_user_error_fn, &png_user_warning_fn);
	png_infop info_ptr = png_create_info_struct(png_ptr);

	png_init_io(png_ptr, file);
	
	//png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
	//png_set_compression_level(png_ptr, Z_BEST_SPEED);
	//png_set_compression_mem_level(png_ptr, 8);
	//png_set_compression_strategy(png_ptr, Z_DEFAULT_STRATEGY);
	//png_set_compression_window_bits(png_ptr, 15);
	//png_set_compression_method(png_ptr, 8);
	//png_set_compression_buffer_size(png_ptr, 8192);

	png_set_IHDR(png_ptr, info_ptr, cx, cy, 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	// low-level write

	png_write_info(png_ptr, info_ptr);
	png_byte *row = new png_byte[cx*4];
	for (int j = 0; j < cy; j++)
	{
		CopyBGRAtoRGB(row, GetScanLine(j), cx);
		png_write_row(png_ptr, row);
	}
	delete row;
	png_write_end(png_ptr, info_ptr);

	png_destroy_write_struct(&png_ptr, &info_ptr);
	fclose(file);
	return true;
}


byte *Bitmap::GetScanLine(int y)
{
	if (y < 0 || y >= cy)
		throw L"Error call GetScanLine (y out of range)";
	return pBitmap + 4*cx*(cy - y - 1);
}

COLORREF Bitmap::GetPixel(int x, int y)
{
	byte *p = pBitmap + 4*(cx*(cy - y - 1) + x);
	return RGB(*(p + 2), *(p + 1), *(p + 0));
}

void Bitmap::SetPixel(int x, int y, COLORREF color)
{
	byte *p = pBitmap + 4*(cx*(cy - y - 1) + x);
	*p++ = GetBValue(color);
	*p++ = GetGValue(color);
	*p   = GetRValue(color);
}

void Bitmap::GetColor(int x, int y, Color *color)
{
	byte *p = pBitmap + 4*(cx*(cy - y - 1) + x);
	color->b = (*p++)/(REAL)255;
	color->g = (*p++)/(REAL)255;
	color->r = (*p  )/(REAL)255;
}

void Bitmap::SetColor(int x, int y, const Color &color)
{
	byte *p = pBitmap + 4*(cx*(cy - y - 1) + x);
	if (color.b >= 1.0f)
		*p++ = 255;
	else if (color.b <= 0.0f)
		*p++ = 0;
	else
		*p++ = (byte)(color.b*255);
	if (color.g >= 1.0f)
		*p++ = 255;
	else if (color.g <= 0.0f)
		*p++ = 0;
	else
		*p++ = (byte)(color.g*255);
	if (color.r >= 1.0f)
		*p = 255;
	else if (color.r <= 0.0f)
		*p = 0;
	else
		*p = (byte)(color.r*255);
}

void Bitmap::GetColorTile(int x, int y, Color *color)
{
	x = x%GetWidth();
	if (x < 0)
		x = GetWidth() + x;
	y = y%GetHeight();
	if (y < 0)
		y = GetHeight() + y;
	GetColor(x, y, color);
}

void Bitmap::SetColorTile(int x, int y, const Color &color)
{
	x = x%GetWidth();
	if (x < 0)
		x = GetWidth() + x;
	y = y%GetHeight();
	if (y < 0)
		y = GetHeight() + y;
	SetColor(x, y, color);
}

void Bitmap::DrawLine(int x1, int y1, int x2, int y2, COLORREF color)
{
	SetDCPenColor(hdcBitmap, color);
	MoveToEx(hdcBitmap, x1, y1, NULL);
	LineTo(hdcBitmap, x2, y2);
}

void Bitmap::DrawText(wchar *text, int i, int j, COLORREF color)
{
	SetTextColor(hdcBitmap, color);
	TextOut(hdcBitmap, i, j, text, (int)wcslen(text));
}
