#include "material.h"

Material MaterialDefault;


Texture::Texture(TextureType type)
{
	this->type = type;
	offset = Vector(0.0f, 0.0f, 0.0f);
	tiling = Vector(1.0f, 1.0f, 1.0f);
	invert = false;
}

Texture::Texture(Texture* tex)
{
	type = tex->type;
	name = tex->name;
	offset = tex->offset;
	tiling = tex->tiling;
	invert = tex->invert;
}

Texture::~Texture()
{
}

Texture *Texture::Copy()
{
	switch (type)
	{
	case ttBitmap:
		return new TextureBitmap((TextureBitmap*)this);
	case ttChessboard:
		return new TextureChessboard((TextureChessboard*)this);
	case ttPerlinNoise:
		return new TexturePerlinNoise((TexturePerlinNoise*)this);
	default:
		throw L"Unknown texture type in Texture::Copy()";
	}
}

void Texture::EvaluteBump(const Vector &v, Vector *bump)
{
	Color color1, color2;

	Vector v11(v.x - 1, v.y, v.z), v12(v.x + 1, v.y, v.z);
	GetColor(v11, &color1);
	GetColor(v12, &color2);
	bump->x = (color2.GetIntens() - color1.GetIntens());

	Vector v21(v.x, v.y - 1, v.z), v22(v.x, v.y + 1, v.z);
	GetColor(v21, &color1);
	GetColor(v22, &color2);
	bump->y = (color2.GetIntens() - color1.GetIntens());
}

void Texture::GetColor(const Vector &v, Color *color)
{
	Vector realV(offset.x + v.x*tiling.x, offset.y + v.y*tiling.y, offset.z + v.z*tiling.z);
	EvaluteColor(realV, color);
	if (invert)
		color->Invert();
}

void Texture::GetBump(const Vector &v, Vector *bump)
{
	Vector realV(offset.x + v.x*tiling.x, offset.y + v.y*tiling.y, offset.z + v.z*tiling.z);
	EvaluteBump(realV, bump);
	if (invert)
		bump->Invert();
}


TextureBitmap::TextureBitmap(wchar *fileName): Texture(ttBitmap), bitmap(fileName)
{
	this->fileName = fileName;
}

TextureBitmap::TextureBitmap(TextureBitmap* tex): Texture(tex), bitmap(tex->fileName.GetString())
{
	this->fileName = tex->fileName;
}

TextureBitmap::~TextureBitmap()
{
}

void TextureBitmap::EvaluteColor(const Vector &v, Color *color)
{
	int ix = (int)FLOOR(v.x*bitmap.GetWidth()), iy = (int)FLOOR(v.y*bitmap.GetHeight());
	// ,          y
	bitmap.GetColorTile(ix, - iy, color);
}

REAL TextureBitmap::Interpolate(REAL v1, REAL v2, REAL t)
{
	//t = (1 - COS(t*PI))/2;
	return v1*(1 - t) + v2*t;
}

void TextureBitmap::EvaluteBump(const Vector &v, Vector *bump)
{
	REAL x = v.x*bitmap.GetWidth();
	REAL y = v.y*bitmap.GetHeight();
	int ix = (int)FLOOR(x), iy = (int)FLOOR(y);
	REAL fx = x - ix, fy = y - iy;

	Color color01, color02;
	Color color10, color11, color12, color13;
	Color color20, color21, color22, color23;
	Color color31, color32;

	// ,          y

	bitmap.GetColorTile(ix - 1, - iy - 0, &color01);
	bitmap.GetColorTile(ix - 1, - iy - 1, &color02);

	bitmap.GetColorTile(ix + 0, - iy + 1, &color10);
	bitmap.GetColorTile(ix + 0, - iy - 0, &color11);
	bitmap.GetColorTile(ix + 0, - iy - 1, &color12);
	bitmap.GetColorTile(ix + 0, - iy - 2, &color13);

	bitmap.GetColorTile(ix + 1, - iy + 1, &color20);
	bitmap.GetColorTile(ix + 1, - iy - 0, &color21);
	bitmap.GetColorTile(ix + 1, - iy - 1, &color22);
	bitmap.GetColorTile(ix + 1, - iy - 2, &color23);

	bitmap.GetColorTile(ix + 2, - iy - 0, &color31);
	bitmap.GetColorTile(ix + 2, - iy - 1, &color32);

	REAL x11 = color11.GetIntens() - color01.GetIntens();
	REAL x21 = color21.GetIntens() - color11.GetIntens();
	REAL x12 = color12.GetIntens() - color02.GetIntens();
	REAL x22 = color22.GetIntens() - color12.GetIntens();
	REAL x1 = Interpolate(x11, x21, fx);
	REAL x2 = Interpolate(x12, x22, fx);
	bump->x = Interpolate(x1, x2, fy);


	REAL y11 = color11.GetIntens() - color10.GetIntens();
	REAL y12 = color12.GetIntens() - color11.GetIntens();
	REAL y21 = color21.GetIntens() - color20.GetIntens();
	REAL y22 = color22.GetIntens() - color21.GetIntens();
	REAL y1 = Interpolate(y11, y12, fy);
	REAL y2 = Interpolate(y21, y22, fy);
	bump->y = Interpolate(y1, y2, fx);

	bump->z = 0.0f;
}

TextureChessboard::TextureChessboard(): Texture(ttChessboard)
{
	evenColor = Color(0.0f, 0.0f, 0.0f);
	oddColor = Color(1.0f, 1.0f, 1.0f);
}

TextureChessboard::TextureChessboard(TextureChessboard* tex): Texture(tex)
{
	evenColor = tex->evenColor;
	oddColor = tex->oddColor;
}

void TextureChessboard::EvaluteColor(const Vector &v, Color *color)
{
	int ix = (int)FLOOR((v.x*2));
	int iy = (int)FLOOR((v.y*2));
	int iz = (int)FLOOR((v.z*2));

	if (abs(ix + iy + iz) & 1)
		*color = oddColor;
	else
		*color = evenColor;
}


TexturePerlinNoise::TexturePerlinNoise(REAL amplitude, REAL frequency, REAL persistence, int octaves): Texture(ttPerlinNoise)
{
	this->amplitude = amplitude;
	this->frequency = frequency;
	this->persistence = persistence;
	this->octaves = octaves;
	baseColor = Color(0.0, 0.0, 0.0);
	noiseColor = Color(1.0, 1.0, 1.0);
}

TexturePerlinNoise::TexturePerlinNoise(TexturePerlinNoise* tex): Texture(tex)
{
	amplitude = tex->amplitude;
	frequency = tex->frequency;
	persistence = tex->persistence;
	octaves = tex->octaves;
	baseColor = tex->baseColor;
	noiseColor = tex->noiseColor;
}

REAL TexturePerlinNoise::Noise(int x, int y, int z, int t)
{
	int v = (x*57)^(y*333)^(z);
	return ((v*(v*v*15731 + 789221) + 1376312589)&0x7fffffff)/(REAL)0x7fffffff;
}

REAL TexturePerlinNoise::SmoothedNoise(int x, int y, int z, int t)
{
	return Noise(x, y, z, t);
}

REAL TexturePerlinNoise::Interpolate(REAL v1, REAL v2, REAL t)
{
	t = (1 - COS(t*PI))/2;
	return v1*(1 - t) + v2*t;
}

REAL TexturePerlinNoise::FinalNoise(REAL x, REAL y, REAL z, int t)
{
	int ix = (int)FLOOR(x), iy = (int)FLOOR(y), iz = (int)FLOOR(z);
	REAL fx = x - ix, fy = y - iy, fz = z - iz;

	REAL n000 = SmoothedNoise(ix, iy, iz, t);
	REAL n001 = SmoothedNoise(ix, iy, iz + 1, t);
	REAL n010 = SmoothedNoise(ix, iy + 1, iz, t);
	REAL n011 = SmoothedNoise(ix, iy + 1, iz + 1, t);
	REAL n100 = SmoothedNoise(ix + 1, iy, iz, t);
	REAL n101 = SmoothedNoise(ix + 1, iy, iz + 1, t);
	REAL n110 = SmoothedNoise(ix + 1, iy + 1, iz, t);
	REAL n111 = SmoothedNoise(ix + 1, iy + 1, iz + 1, t);
	
	REAL i00 = Interpolate(n000, n001, fz);
	REAL i01 = Interpolate(n010, n011, fz);
	REAL i10 = Interpolate(n100, n101, fz);
	REAL i11 = Interpolate(n110, n111, fz);

	REAL i0 = Interpolate(i00, i01, fy);
	REAL i1 = Interpolate(i10, i11, fy);

	REAL i = Interpolate(i0, i1, fx);

	return i;
}

void TexturePerlinNoise::EvaluteColor(const Vector &v, Color *color)
{
	REAL freq = frequency;
	REAL amp = amplitude;
	REAL total = 0;
	for (int i = 0; i < octaves; i++)
	{
		total += amp*FinalNoise(v.x*freq, v.y*freq, v.z*freq, 13);
		freq *= 2;
		amp *= persistence;
	}
	if (total > 1.0f)
		total = 1.0f;

	//total = (1 + COS(PI*total))/2;

	color->r = baseColor.r + (noiseColor.r - baseColor.r)*total;
	color->g = baseColor.g + (noiseColor.g - baseColor.g)*total;
	color->b = baseColor.b + (noiseColor.b - baseColor.b)*total;
}


Material::Material()
{
	ambient = Color(0.0f, 0.0f, 0.0f);
	diffuse = Color(Randomf(0.0f, 1.0f), Randomf(0.0f, 1.0f), Randomf(0.0f, 1.0f));
	specular = Color(1.0f, 1.0f, 1.0f);
	opacity = Color(1.0f, 1.0f, 1.0f);
	reflection = Color(0.0f, 0.0f, 0.0f);

	specularAmount = 10.0f;
	specularLevel = 1.0f;
	refractionLevel = 1.0f;
	bumpAmount = 1.0f;

	ambientMap = NULL;
	diffuseMap = NULL;
	specularMap = NULL;
	opacityMap = NULL;
	reflectionMap = NULL;
	bumpMap = NULL;
}

Material::~Material()
{
	delete ambientMap;
	delete diffuseMap;
	delete specularMap;
	delete opacityMap;
	delete reflectionMap;
	delete bumpMap;
}

Material *Material::Copy()
{
	Material *mat = new Material();
	
	mat->name = name;
	mat->ambient = ambient;
	mat->diffuse = diffuse;
	mat->specular = specular;
	mat->opacity = opacity;
	mat->reflection = reflection;

	if (ambientMap)
		mat->ambientMap = ambientMap->Copy();
	if (diffuseMap)
		mat->diffuseMap = diffuseMap->Copy();
	if (specularMap)
		mat->specularMap = specularMap->Copy();
	if (opacityMap)
		mat->opacityMap = opacityMap->Copy();
	if (reflectionMap)
		mat->reflectionMap = reflectionMap->Copy();
	if (bumpMap)
		mat->bumpMap = bumpMap->Copy();

	mat->specularAmount = specularAmount;
	mat->specularLevel = specularLevel;
	mat->refractionLevel = refractionLevel;
	mat->bumpAmount = bumpAmount;

	return mat;
}
