#include "nodes.h"
#include "scene.h"


Node::Node(Scene *scene, Group *parent)
{
	this->scene = scene;
	this->parent = parent;
	if (parent)
		parent->Add(this);

	MakeIdentity(&mTransform);
	mTransformAnimation = NULL;
}

Node::~Node()
{
	if (parent)
		parent->Remove(this);
	if (mTransformAnimation)
		delete mTransformAnimation;
}

void Node::Prepare(const Matrix &m, bool render, TriangleArray *list)
{
	if (mTransformAnimation)
		mTransform = mTransformAnimation[scene->GetCurrentFrame()];
	MultMatrices(&mTransformWorld, mTransform, m);
}


Group::Group(Scene *scene, Group *parent): Node(scene, parent)
{
}

Group::~Group()
{
	for (int i = subNodes.GetCount() - 1; i >= 0; i--)
		delete subNodes[i];
}

void Group::Prepare(const Matrix &m, bool render, PTriangleArray list)
{
	Node::Prepare(m, render, list);
	for (int i = 0; i < subNodes.GetCount(); i++)
		subNodes[i]->Prepare(mTransformWorld, render, list);
}


Mesh::Mesh(Scene *scene, Group *parent): Node(scene, parent)
{
	matRef = MatRefDefault;
	mat = NULL;
}

void Mesh::Prepare(const Matrix &m, bool render, TriangleArray *list)
{
	Node::Prepare(m, render, list);

	Material *mat;
	if (matRef != MatRefDefault && scene)
		mat = scene->materials[matRef];
	else if (this->mat)
		mat = this->mat;
	else if (matRef == MatRefDefault)
		mat = &MaterialDefault;

	for (int i = 0; i < triangles.GetCount(); i++)
	{
		Triangle *triangle = triangles[i];
		triangle->material = mat;

		triangle->PrepareWorldCoord(mTransformWorld);
		list->Add(triangle);
		triangle->Prepare();
	}
}

void Mesh::SetMatRef(int matRef)
{
	this->matRef = matRef;
}

void Mesh::SetMat(Material *mat)
{
	this->mat = mat;
}


Light::Light(Scene *scene, Group *parent): Node(scene, parent), view(NULL, NULL)
{
	color = Color(1.0f, 1.0f, 1.0f);
	intens = 1.0f;
	type = ltOmni;
	decay = ldInverse;
	decayStart = 0.0f;
	coneHotspot = 30.0f;
	coneFalloff = 45.0f;

	Vector 
		pt0(0.0f, 0.0f, - 1.0f), pt1(0.0f, 0.0f, 1.0f), 
		pt2(1.0f, 1.0f, 0.0f), pt3(- 1.0f, 1.0f, 0.0f), 
		pt4(- 1.0f, - 1.0f, 0.0f), pt5(1.0f, - 1.0f, 0.0f);
	Triangle *tri;

	view.SetMat(&matView);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt2; tri->pt2 = pt3; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt1; tri->pt1 = pt2; tri->pt2 = pt3; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt3; tri->pt2 = pt4; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt1; tri->pt1 = pt3; tri->pt2 = pt4; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt4; tri->pt2 = pt5; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt1; tri->pt1 = pt4; tri->pt2 = pt5; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt5; tri->pt2 = pt2; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt1; tri->pt1 = pt5; tri->pt2 = pt2; view.Add(tri);
}

void Light::Prepare(const Matrix &m, bool render, TriangleArray *list)
{
	Node::Prepare(m, render, list);
	if (!render)
	{
		matView.diffuse = Color(1.0f, 1.0f, 0.0f);
		view.Prepare(mTransformWorld, render, list);
	}
	TransformVector(&dirw, Vector(0.0f, 0.0f, 1.0f), mTransformWorld);

	cosHotspot = COS(DegreeToRadian(coneHotspot/2));
	cosFalloff = COS(DegreeToRadian(coneFalloff/2));
}

void Light::EvaluteDir(const Vector &pt, Vector *dir)
{
	switch (type)
	{
		case ltOmni:
		case ltSpot:
		{
			dir->x = GetX() - pt.x;
			dir->y = GetY() - pt.y;
			dir->z = GetZ() - pt.z;
			break;
		}
		case ltDirectional:
		{
			dir->x = 100.0f*dirw.x;
			dir->y = 100.0f*dirw.y;
			dir->z = 100.0f*dirw.z;
			break;
		}
	}
}

void Light::EvaluteLight(const Vector &pt, const Vector &dir, Color *color)
{
	switch (type)
	{
		case ltOmni:
		{
			color->r = this->color.r;
			color->g = this->color.g;
			color->b = this->color.b;
			break;
		}
		case ltSpot:
		{
			REAL cos = ScalarProductNorm(dirw, dir);
			if (cos >= cosHotspot)
			{
				color->r = this->color.r;
				color->g = this->color.g;
				color->b = this->color.b;
			}
			else if (cos > cosFalloff)
			{
				REAL kHotspot = (cos - cosFalloff)/(cosHotspot - cosFalloff);
				color->r = this->color.r*kHotspot;
				color->g = this->color.g*kHotspot;
				color->b = this->color.b*kHotspot;
			}
			break;
		}
		case ltDirectional:
		{
			color->r = this->color.r;
			color->g = this->color.g;
			color->b = this->color.b;
			break;
		}
	}
}


Camera::Camera(Scene *scene, Group *parent): Node(scene, parent), view(NULL, NULL)
{
	Vector pt0(0.0f, 0.0f, 0.0f), 
		pt1(+ 1.0f, + 1.0f, - 2.0f), pt2(- 1.0f, + 1.0f, - 2.0f), 
		pt3(- 1.0f, - 1.0f, - 2.0f), pt4(+ 1.0f, - 1.0f, - 2.0f);
	Triangle *tri;

	view.SetMat(&matView);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt1; tri->pt2 = pt2; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt2; tri->pt2 = pt3; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt3; tri->pt2 = pt4; view.Add(tri);
	tri = new Triangle(); tri->pt0 = pt0; tri->pt1 = pt4; tri->pt2 = pt1; view.Add(tri);
}

void Camera::Prepare(const Matrix &m, bool render, TriangleArray *list)
{
	Node::Prepare(m, render, list);
	if (!render)
	{
		matView.diffuse = Color(0.0f, 0.0f, 1.0f);
		view.Prepare(mTransformWorld, render, list);
	}
}
