#include "scene.h"


Scene::Scene(DWORD nWidth, DWORD nHeight, REAL angleView): root(NULL, NULL)
{
	this->nWidth = nWidth;
	this->nHeight = nHeight;
	CreateBuffer();

	zDrawClipNear = - 10.0f;	
	callbackFunc = NULL;	
	currentFrame = 0;
	totalFrame = 0;
	cubeCount = 32;
	root.scene = this;

	SetAngleView(angleView);
	MakeIdentity(&mView);
}

Scene::~Scene()
{
	for (int i = lights.GetCount() - 1; i >= 0; i--)
		delete lights[i];
	for (i = materials.GetCount() - 1; i >= 0; i--)
		delete materials[i];
	DeleteBuffer();
}

void Scene::Callback(SceneState state, int done, int total, bool *cancel)
{
	if (callbackFunc)
		callbackFunc(state, done, total, cancel);
}

void Scene::CreateBuffer()
{
	buffer = new Bitmap(nWidth, nHeight);
}

void Scene::DeleteBuffer()
{
	delete buffer;
}

void Scene::SetAngleView(REAL value)
{
	if (value <= 0 || value >= PI)
		throw L"Angle view must be in (0..180) range";
	angleView = value;
	zProjection = - COS(angleView/2)/SIN(angleView/2);
}

void Scene::ViewToScreen(const Vector &v, REAL *x, REAL *y)
{
	REAL size = (REAL)max(nWidth, nHeight);
	*x = (nWidth + size*zProjection*v.x/v.z)/2;
	*y = (nHeight - size*zProjection*v.y/v.z)/2;
}

void Scene::ScreenToView(Vector *res, REAL x, REAL y)
{
	REAL size = (REAL)max(nWidth, nHeight);
	res->x = (2*x - nWidth)/size;
	res->y = (nHeight - 2*y)/size;
	res->z = zProjection;
}

void Scene::WorldToScreen(const Vector &v, REAL *x, REAL *y)
{
	Vector res;
	Transform(&res, v, mView);
	ViewToScreen(res, x, y);
}

void Scene::ScreenToWorld(Vector *res, REAL x, REAL y)
{
	Vector v;
	ScreenToView(&v, x, y);
	Transform(res, v, mViewReverse);
}

void Scene::ViewToScreeni(const Vector &v, int *ix, int *iy)
{
	REAL x, y;
	ViewToScreen(v, &x, &y);
	*ix = (int)FLOOR(x);
	*iy = (int)FLOOR(y);
}

void Scene::ScreenToViewi(Vector *res, int ix, int iy)
{
	ScreenToView(res, (REAL)ix, (REAL)iy);
}

void Scene::WorldToScreeni(const Vector &v, int *ix, int *iy)
{
	REAL x, y;
	ViewToScreen(v, &x, &y);
	*ix = (int)FLOOR(x);
	*iy = (int)FLOOR(y);
}

void Scene::ScreenToWorldi(Vector *res, int ix, int iy)
{
	ScreenToWorld(res, (REAL)ix, (REAL)iy);
}


void Scene::ClearBuffer(COLORREF color)
{
	SetDCPenColor(buffer->hdcBitmap, color);
	SetDCBrushColor(buffer->hdcBitmap, color);
	Rectangle(buffer->hdcBitmap, 0, 0, buffer->GetWidth(), buffer->GetHeight());
	//ZeroMemory(buffer->pBitmap, nWidth*nHeight*4);
}


void Scene::DrawLine(const Vector &p1, const Vector &p2, COLORREF color)
{
	Vector view1, view2;
	Transform(&view1, p1, mView);
	Transform(&view2, p2, mView);

	if (view1.z > zDrawClipNear && view2.z > zDrawClipNear)
		return;

	if (view1.z > zDrawClipNear)
	{
		REAL t = (zDrawClipNear - view2.z)/(view1.z - view2.z);
		view1.x = view2.x + (view1.x - view2.x)*t;
		view1.y = view2.y + (view1.y - view2.y)*t;
		view1.z = zDrawClipNear;
	}
	else if (view2.z > zDrawClipNear)
	{
		REAL t = (zDrawClipNear - view1.z)/(view2.z - view1.z);
		view2.x = view1.x + (view2.x - view1.x)*t;
		view2.y = view1.y + (view2.y - view1.y)*t;
		view2.z = zDrawClipNear;
	}

	int x1, y1, x2, y2;
	ViewToScreeni(view1, &x1, &y1);
	ViewToScreeni(view2, &x2, &y2);
	buffer->DrawLine(x1, y1, x2, y2, color);
}

void Scene::DrawVector(const Vector &p, const Vector &dir, COLORREF color)
{
	DrawLine(p, Vector(p.x + dir.x, p.y + dir.y, p.z + dir.z), color);
}

void Scene::DrawAxis(REAL size)
{
	Vector o(0, 0, 0);
	Vector ox(size, 0, 0);
	Vector oy(0, size, 0);
	Vector oz(0, 0, size);
	DrawLine(o, ox, RGB(255, 0, 0));
	DrawLine(o, oy, RGB(0, 255, 0));
	DrawLine(o, oz, RGB(0, 0, 255));
}

void Scene::DrawTriangle(const Triangle &t, COLORREF color)
{
	DrawLine(t.ptw0, t.ptw1, color);
	DrawLine(t.ptw1, t.ptw2, color);
	DrawLine(t.ptw2, t.ptw0, color);
}

void Scene::DrawCube(int ix, int iy, int iz, COLORREF color)
{
	Vector p000(GetCubeX(ix), GetCubeY(iy), GetCubeZ(iz));
	Vector p100(GetCubeX(ix + 1), GetCubeY(iy), GetCubeZ(iz));
	Vector p010(GetCubeX(ix), GetCubeY(iy + 1), GetCubeZ(iz));
	Vector p110(GetCubeX(ix + 1), GetCubeY(iy + 1), GetCubeZ(iz));
	Vector p001(GetCubeX(ix), GetCubeY(iy), GetCubeZ(iz + 1));
	Vector p101(GetCubeX(ix + 1), GetCubeY(iy), GetCubeZ(iz + 1));
	Vector p011(GetCubeX(ix), GetCubeY(iy + 1), GetCubeZ(iz + 1));
	Vector p111(GetCubeX(ix + 1), GetCubeY(iy + 1), GetCubeZ(iz + 1));

	DrawLine(p000, p001, color);
	DrawLine(p010, p011, color);
	DrawLine(p100, p101, color);
	DrawLine(p110, p111, color);

	DrawLine(p000, p010, color);
	DrawLine(p001, p011, color);
	DrawLine(p100, p110, color);
	DrawLine(p101, p111, color);

	DrawLine(p000, p100, color);
	DrawLine(p001, p101, color);
	DrawLine(p010, p110, color);
	DrawLine(p011, p111, color);
}

void Scene::DrawCube(const Vector &ptw0, const Vector &ptw1, COLORREF color)
{
	Vector p000(ptw0.x, ptw0.y, ptw0.z);
	Vector p100(ptw1.x, ptw0.y, ptw0.z);
	Vector p010(ptw0.x, ptw1.y, ptw0.z);
	Vector p110(ptw1.x, ptw1.y, ptw0.z);
	Vector p001(ptw0.x, ptw0.y, ptw1.z);
	Vector p101(ptw1.x, ptw0.y, ptw1.z);
	Vector p011(ptw0.x, ptw1.y, ptw1.z);
	Vector p111(ptw1.x, ptw1.y, ptw1.z);

	DrawLine(p000, p001, color);
	DrawLine(p010, p011, color);
	DrawLine(p100, p101, color);
	DrawLine(p110, p111, color);

	DrawLine(p000, p010, color);
	DrawLine(p001, p011, color);
	DrawLine(p100, p110, color);
	DrawLine(p101, p111, color);

	DrawLine(p000, p100, color);
	DrawLine(p001, p101, color);
	DrawLine(p010, p110, color);
	DrawLine(p011, p111, color);
}


void Scene::Render()
{
	ClearBuffer(backgroundStatic.GetRGB());

	ReverseMatrix(&mViewReverse, mView);

	TriangleArray list;
	Matrix ident;
	MakeIdentity(&ident);
	root.Prepare(ident, false, &list);

	//CreateBoundingBox(&list);

	DrawAxis(10.0f);

	//DrawCube(minBounding, maxBounding, RGB(255, 255, 255));

	for (int i = 0; i < list.GetCount(); i++)
	{
		PTriangle tri = list[i];
		DrawTriangle(*tri, tri->material->diffuse.GetRGB());
	}
}

void Scene::SetCubeCount(int value)
{
	if (value < 1 || value > MAX_CUBE_COUNT)
		throw L"Error cube count";
	if (value != cubeCount)
	{
		cubeCount = value;
	}
}


void Scene::CreateBoundingBox(PTriangleArray list)
{
	Vector vmin(INFINITY, INFINITY, INFINITY);
	Vector vmax(- INFINITY, - INFINITY, - INFINITY);

	for (int i = 0; i < list->GetCount(); i++)
	{
		PTriangle t = (*list)[i];
		if (t->ptw0.x < vmin.x) vmin.x = t->ptw0.x;
		if (t->ptw0.y < vmin.y) vmin.y = t->ptw0.y;
		if (t->ptw0.z < vmin.z) vmin.z = t->ptw0.z;
		if (t->ptw0.x > vmax.x) vmax.x = t->ptw0.x;
		if (t->ptw0.y > vmax.y) vmax.y = t->ptw0.y;
		if (t->ptw0.z > vmax.z) vmax.z = t->ptw0.z;

		if (t->ptw1.x < vmin.x) vmin.x = t->ptw1.x;
		if (t->ptw1.y < vmin.y) vmin.y = t->ptw1.y;
		if (t->ptw1.z < vmin.z) vmin.z = t->ptw1.z;
		if (t->ptw1.x > vmax.x) vmax.x = t->ptw1.x;
		if (t->ptw1.y > vmax.y) vmax.y = t->ptw1.y;
		if (t->ptw1.z > vmax.z) vmax.z = t->ptw1.z;

		if (t->ptw2.x < vmin.x) vmin.x = t->ptw2.x;
		if (t->ptw2.y < vmin.y) vmin.y = t->ptw2.y;
		if (t->ptw2.z < vmin.z) vmin.z = t->ptw2.z;
		if (t->ptw2.x > vmax.x) vmax.x = t->ptw2.x;
		if (t->ptw2.y > vmax.y) vmax.y = t->ptw2.y;
		if (t->ptw2.z > vmax.z) vmax.z = t->ptw2.z;
	}

	minBounding = Vector(vmin.x - EPS_BOUNDING, vmin.y - EPS_BOUNDING, vmin.z - EPS_BOUNDING);
	maxBounding = Vector(vmax.x + EPS_BOUNDING, vmax.y + EPS_BOUNDING, vmax.z + EPS_BOUNDING);
	
	dBounding.x = (maxBounding.x - minBounding.x)/cubeCount;
	dBounding.y = (maxBounding.y - minBounding.y)/cubeCount;
	dBounding.z = (maxBounding.z - minBounding.z)/cubeCount;

	dBoundingReverse.x = cubeCount/(maxBounding.x - minBounding.x);
	dBoundingReverse.y = cubeCount/(maxBounding.y - minBounding.y);
	dBoundingReverse.z = cubeCount/(maxBounding.z - minBounding.z);
}

void Scene::CreateCubes(PTriangleArray list)
{
	CreateBoundingBox(list);

	int i, j, k;
	for (i = 0; i < cubeCount; i++)
		for (j = 0; j < cubeCount; j++)
			for (k = 0; k < cubeCount; k++)
				cubes[i][j][k] = NULL;

	//     
	//       
	int triangleindex;
	PTriangle ptri;
	REAL t;

	for (triangleindex = 0; triangleindex < list->GetCount(); triangleindex++)
	{
		ptri = (*list)[triangleindex];

		Vector p0 = ptri->ptw0, p1 = ptri->ptw1, p2 = ptri->ptw2;
		Vector dir01, dir12, dir20;

		VectorSubtraction(&dir01, p1, p0);
		VectorSubtraction(&dir12, p2, p1);
		VectorSubtraction(&dir20, p0, p2);

		CubeInfo cube;
		CubeInfo cube01;
		CubeInfo cube12;
		CubeInfo cube20;

		if (!FirstCube(&cube01, &p0, dir01))
			throw L"Error in CreateCubes()!";
		if (!FirstCube(&cube12, &p1, dir12))
			throw L"Error in CreateCubes()!";
		if (!FirstCube(&cube20, &p2, dir20))
			throw L"Error in CreateCubes()!";

		cube = cube01;
		AddOnceTriToCube(cube.ix, cube.iy, cube.iz, ptri);
		while (NextCube(&cube, &p0, dir01) && 
			(cube.ix != cube12.ix || cube.iy != cube12.iy || cube.iz != cube12.iz))
			AddOnceTriToCube(cube.ix, cube.iy, cube.iz, ptri);

		cube = cube12;
		AddOnceTriToCube(cube.ix, cube.iy, cube.iz, ptri);
		while (NextCube(&cube, &p1, dir12) && 
			(cube.ix != cube20.ix || cube.iy != cube20.iy || cube.iz != cube20.iz))
			AddOnceTriToCube(cube.ix, cube.iy, cube.iz, ptri);

		cube = cube20;
		AddOnceTriToCube(cube.ix, cube.iy, cube.iz, ptri);
		while (NextCube(&cube, &p2, dir20) && 
			(cube.ix != cube01.ix || cube.iy != cube01.iy || cube.iz != cube01.iz))
			AddOnceTriToCube(cube.ix, cube.iy, cube.iz, ptri);

		int xmin = Min(cube01.ix, cube12.ix, cube20.ix);
		int ymin = Min(cube01.iy, cube12.iy, cube20.iy);
		int zmin = Min(cube01.iz, cube12.iz, cube20.iz);
		int xmax = Max(cube01.ix, cube12.ix, cube20.ix);
		int ymax = Max(cube01.iy, cube12.iy, cube20.iy);
		int zmax = Max(cube01.iz, cube12.iz, cube20.iz);

		for (i = xmin; i <= xmax; i++)
			for (j = ymin; j <= ymax; j++)
			{
				Vector p(GetCubeX(i), GetCubeY(j), minBounding.z);
				Vector dir(0, 0, 1);
				if (ptri->Intersect(p, dir, &t))
				{
					k = GetCubeIZ(p.z + dir.z*t);
					AddOnceTriToCube(i, j, k, ptri);
					AddOnceTriToCube(i - 1, j, k, ptri);
					AddOnceTriToCube(i, j - 1, k, ptri);
					AddOnceTriToCube(i - 1, j - 1, k, ptri);
				}
			}

		for (j = ymin; j <= ymax; j++)
			for (k = zmin; k <= zmax; k++)
			{
				Vector p(minBounding.x, GetCubeY(j), GetCubeZ(k));
				Vector dir(1, 0, 0);
				if (ptri->Intersect(p, dir, &t))
				{
					i = GetCubeIX(p.x + dir.x*t);
					AddOnceTriToCube(i, j, k, ptri);
					AddOnceTriToCube(i, j, k - 1, ptri);
					AddOnceTriToCube(i, j - 1, k, ptri);
					AddOnceTriToCube(i, j - 1, k - 1, ptri);
				}
			}

		for (k = zmin; k <= zmax; k++)
			for (i = xmin; i <= xmax; i++)
			{
				Vector p(GetCubeX(i), minBounding.y, GetCubeZ(k));
				Vector dir(0, 1, 0);
				if (ptri->Intersect(p, dir, &t))
				{
					j = GetCubeIY(p.y + dir.y*t);
					AddOnceTriToCube(i, j, k, ptri);
					AddOnceTriToCube(i, j, k - 1, ptri);
					AddOnceTriToCube(i - 1, j, k, ptri);
					AddOnceTriToCube(i - 1, j, k - 1, ptri);
				}
			}
	}
}

void Scene::AddOnceTriToCube(int i, int j, int k, PTriangle ptri)
{
	if (i < 0 || i >= cubeCount || j < 0 || j >= cubeCount || k < 0 || k >= cubeCount)
		throw L"Error call AddOnceTriToCube (out of range)!";
	if (cubes[i][j][k] == NULL)
		cubes[i][j][k] = new SimpleTriangleArray();
	cubes[i][j][k]->AddOnce(ptri);
}

void Scene::DeleteCubes()
{
	for (int i = 0; i < cubeCount; i++)
		for (int j = 0; j < cubeCount; j++)
			for (int k = 0; k < cubeCount; k++)
				if (cubes[i][j][k])
				{
					delete cubes[i][j][k];
					cubes[i][j][k] = NULL;
				}
}


bool Scene::FirstCube(CubeInfo *cube, Vector *p, const Vector &dir)
{
	Vector v = *p;

	if (v.x < minBounding.x)
	{
		if (dir.x <= 0)
			return false;		
		REAL t = (minBounding.x - v.x)/dir.x + EPS;
		v.x = v.x + dir.x*t;
		v.y = v.y + dir.y*t;
		v.z = v.z + dir.z*t;
	}
	else if (v.x > maxBounding.x)
	{
		if (dir.x >= 0)
			return false;		
		REAL t = (maxBounding.x - v.x)/dir.x + EPS;
		v.x = v.x + dir.x*t;
		v.y = v.y + dir.y*t;
		v.z = v.z + dir.z*t;
	}

	if (v.y < minBounding.y)
	{
		if (dir.y <= 0)
			return false;		
		REAL t = (minBounding.y - v.y)/dir.y + EPS;
		v.x = v.x + dir.x*t;
		v.y = v.y + dir.y*t;
		v.z = v.z + dir.z*t;
	}
	else if (v.y > maxBounding.y)
	{
		if (dir.y >= 0)
			return false;		
		REAL t = (maxBounding.y - v.y)/dir.y + EPS;
		v.x = v.x + dir.x*t;
		v.y = v.y + dir.y*t;
		v.z = v.z + dir.z*t;
	}

	if (v.z < minBounding.z)
	{
		if (dir.z <= 0)
			return false;		
		REAL t = (minBounding.z - v.z)/dir.z + EPS;
		v.x = v.x + dir.x*t;
		v.y = v.y + dir.y*t;
		v.z = v.z + dir.z*t;
	}
	else if (v.z > maxBounding.z)
	{
		if (dir.z >= 0)
			return false;		
		REAL t = (maxBounding.z - v.z)/dir.z + EPS;
		v.x = v.x + dir.x*t;
		v.y = v.y + dir.y*t;
		v.z = v.z + dir.z*t;
	}

	CubeInfo cub;
	cub.ix = GetCubeIX(v.x);
	cub.iy = GetCubeIY(v.y);
	cub.iz = GetCubeIZ(v.z);
	if (cub.ix < 0 || cub.ix >= cubeCount || cub.iy < 0 || cub.iy >= cubeCount || cub.iz < 0 || cub.iz >= cubeCount)
		return false;

	*p = v;
	*cube = cub;
	return true;
}

bool Scene::NextCube(CubeInfo *cube, Vector *p, const Vector &dir)
{
	REAL tx, ty, tz, t;

	if (dir.x > 0)
		tx = (GetCubeX(cube->ix + 1) - p->x)/dir.x;
	else if (dir.x < 0)
		tx = (GetCubeX(cube->ix) - p->x)/dir.x;
	else
		tx = INFINITY;

	if (dir.y > 0)
		ty = (GetCubeY(cube->iy + 1) - p->y)/dir.y;
	else if (dir.y < 0)
		ty = (GetCubeY(cube->iy) - p->y)/dir.y;
	else
		ty = INFINITY;

	if (dir.z > 0)
		tz = (GetCubeZ(cube->iz + 1) - p->z)/dir.z;
	else if (dir.z < 0)
		tz = (GetCubeZ(cube->iz) - p->z)/dir.z;
	else
		tz = INFINITY;

	if (tx < ty && tx < tz)
	{
		t = tx;
		if (dir.x > 0)
			cube->ix++;
		else
			cube->ix--;
		if (cube->ix < 0 || cube->ix >= cubeCount)
			return false;
	}
	else if (ty < tz)
	{
		t = ty;
		if (dir.y > 0)
			cube->iy++;
		else
			cube->iy--;
		if (cube->iy < 0 || cube->iy >= cubeCount)
			return false;
	}
	else if (tz < INFINITY)
	{
		t = tz;
		if (dir.z > 0)
			cube->iz++;
		else
			cube->iz--;
		if (cube->iz < 0 || cube->iz >= cubeCount)
			return false;
	}
	else
		return false;
		//throw L"Error in NextCube(...). All t is infinity!";

	p->x = p->x + dir.x*t;
	p->y = p->y + dir.y*t;
	p->z = p->z + dir.z*t;
	return true;
}


bool Scene::Intersect(const Vector &o, const Vector &dir, REAL *tintersect, PTriangle *ppTri)
{
	Vector p = o;
	CubeInfo cube;

	if (FirstCube(&cube, &p, dir))
	{
		do
		{
			PSimpleTriangleArray ptriangles = cubes[cube.ix][cube.iy][cube.iz];
			PTriangle ptri, ptrimin = NULL;

			REAL t, tmin = INFINITY;
			if (ptriangles != NULL)
			{
				for (int i = 0; i < ptriangles->GetCount(); i++)
				{
					ptri = (*ptriangles)[i];
					if (ptri->Intersect(o, dir, &t) && (t < tmin) && 
						GetCubeIX(ptri->intersectPt.x - EPS) <= cube.ix && GetCubeIX(ptri->intersectPt.x + EPS) >= cube.ix && 
						GetCubeIY(ptri->intersectPt.y - EPS) <= cube.iy && GetCubeIY(ptri->intersectPt.y + EPS) >= cube.iy && 
						GetCubeIZ(ptri->intersectPt.z - EPS) <= cube.iz && GetCubeIZ(ptri->intersectPt.z + EPS) >= cube.iz)
					{
						ptrimin = ptri;
						tmin = t;
					}
				}
			}

			if (ptrimin)
			{
				*tintersect = tmin;
				if (ppTri)
					*ppTri = ptrimin;
				return true;
			}
		}
		while (NextCube(&cube, &p, dir));
	}
	return false;
}


void Scene::RayColor(const Vector &oRay, const Vector &dirRay_, int recurseLevel, Color *outColor)
{
	Vector dirRay = dirRay_;
	dirRay.Normalize();

	REAL tIntersect;
	Triangle *triIntersect;

	if (Intersect(oRay, dirRay, &tIntersect, &triIntersect))
	{
		//   
		Material *mat = triIntersect->material;
		Color ambientMat, diffuseMat, specularMat, opacityMat, reflectionMat;
		Vector bumpMat;

		Vector texIntersect;
		if (mat->ambientMap || mat->diffuseMap || mat->specularMap || mat->opacityMap || mat->reflectionMap || mat->bumpMap)
			triIntersect->GetIntersectTexture(&texIntersect);

		if (mat->ambientMap)
			mat->ambientMap->GetColor(texIntersect, &ambientMat);
		else
			ambientMat = mat->ambient;
		if (mat->diffuseMap)
			mat->diffuseMap->GetColor(texIntersect, &diffuseMat);
		else
			diffuseMat = mat->diffuse;
		if (mat->specularMap)
			mat->specularMap->GetColor(texIntersect, &specularMat);
		else
			specularMat = mat->specular;
		if (mat->opacityMap)
			mat->opacityMap->GetColor(texIntersect, &opacityMat);
		else
			opacityMat = mat->opacity;
		if (mat->reflectionMap)
			mat->reflectionMap->GetColor(texIntersect, &reflectionMat);
		else
			reflectionMat = mat->reflection;

		//  
		Vector n;
		triIntersect->GetIntersectNormalw(&n);
		if (mat->bumpMap)
		{
			mat->bumpMap->GetBump(texIntersect, &bumpMat);
			n.x += mat->bumpAmount*(triIntersect->bump_du.x*bumpMat.x + triIntersect->bump_dv.x*bumpMat.y);
			n.y += mat->bumpAmount*(triIntersect->bump_du.y*bumpMat.x + triIntersect->bump_dv.y*bumpMat.y);
			n.z += mat->bumpAmount*(triIntersect->bump_du.z*bumpMat.x + triIntersect->bump_dv.z*bumpMat.y);
			n.Normalize();
		}

		//  
		Vector dirReflection, vector_n2nd;
		REAL scalar_nd = ScalarProduct(n, dirRay);
		bool isFrontRay = (scalar_nd < 0);
		VectorScale(&vector_n2nd, n, - 2*scalar_nd);
		VectorAddition(&dirReflection, dirRay, vector_n2nd);
        

		//  
		Color pointColor(ambientMat.r*ambientStatic.r, ambientMat.g*ambientStatic.g, ambientMat.b*ambientStatic.b);
		//  
		Color pointSpecularColor;

		for (int i = 0; i < lights.GetCount(); i++)
		{
			Light *light = lights[i];

			Color lightIntens(1.0f, 1.0f, 1.0f);
			Vector dirLight;
			light->EvaluteDir(triIntersect->intersectPt, &dirLight);

			//      
			if (triIntersect->IsFrontPointw(light->GetPos()) == triIntersect->IsFrontPointw(oRay))
			{
				//   ,   
				// ,    
				Vector oLightIntersect = triIntersect->intersectPt;
				REAL tLightIntersect, tLightIntersectAll = 0;
				Triangle *triLightIntersect;
				while (
					Intersect(oLightIntersect, dirLight, &tLightIntersect, &triLightIntersect) && 
					((tLightIntersectAll += tLightIntersect) < 1.0f) && 
					!lightIntens.IsBlack())
				{
					Material *mat = triLightIntersect->material;
					if (mat->opacityMap)
					{
						Vector tex;
						triLightIntersect->GetIntersectTexture(&tex);
						Color col;
						mat->opacityMap->GetColor(tex, &col);
						lightIntens.r *= 1.0f - col.r;
						lightIntens.g *= 1.0f - col.g;
						lightIntens.b *= 1.0f - col.b;
					}
					else
					{
						lightIntens.r *= 1.0f - mat->opacity.r;
						lightIntens.g *= 1.0f - mat->opacity.g;
						lightIntens.b *= 1.0f - mat->opacity.b;
					}
					oLightIntersect = triLightIntersect->intersectPt;
				}

				if (!lightIntens.IsBlack())
				{
					Color lightColor;
					light->EvaluteLight(triIntersect->intersectPt, dirLight, &lightColor);
					dirLight.Normalize();

					REAL cosDiffuse = ScalarProduct(n, dirLight);
					if (cosDiffuse > 0)
					{
						pointColor.r += diffuseMat.r*lightColor.r*cosDiffuse;
						pointColor.g += diffuseMat.g*lightColor.g*cosDiffuse;
						pointColor.b += diffuseMat.b*lightColor.b*cosDiffuse;
					}

					REAL cosSpecular = ScalarProduct(dirReflection, dirLight);
					if (cosSpecular > 0)
					{
						REAL cosSpecularPow = mat->specularAmount*POW(cosSpecular, SPECULARLEVEL_MIN + mat->specularLevel*SPECULARLEVEL_SCALE);
						pointSpecularColor.r += specularMat.r*lightColor.r*cosSpecularPow;
						pointSpecularColor.g += specularMat.g*lightColor.g*cosSpecularPow;
						pointSpecularColor.b += specularMat.b*lightColor.b*cosSpecularPow;
					}
				}
			}
		}


		Color rayReflectionColor, rayRefractionColor;

		// reflection
		if (recurseLevel > 0 && !reflectionMat.IsBlack())
		{
			RayColor(triIntersect->intersectPt, dirReflection, recurseLevel - 1, &rayReflectionColor);
		}

		// refraction
		if (recurseLevel > 0 && !opacityMat.IsWhite())
		{
			REAL k_refraction = isFrontRay? 1/mat->refractionLevel: mat->refractionLevel;
			Vector dirRefraction, vector_nd, vector_nnd, vector_knnd, vector_kn;
			VectorProduct(&vector_nd, n, dirRay);
			VectorProduct(&vector_nnd, n, vector_nd);
			VectorScale(&vector_knnd, vector_nnd, - k_refraction);
			REAL kn2 = 1 - vector_knnd.LengthSq();
			bool isFullReflection = kn2 < 0;
			if (!isFullReflection)
			{
				VectorScale(&vector_kn, n, isFrontRay? - SQRT(kn2): SQRT(kn2));
				VectorAddition(&dirRefraction, vector_knnd, vector_kn);
				RayColor(triIntersect->intersectPt, dirRefraction, recurseLevel - 1, &rayRefractionColor);
			}
		}

		//  
		outColor->r = 
			(pointColor.r*opacityMat.r + rayRefractionColor.r*(1 - opacityMat.r))*(1 - reflectionMat.r) + 
			rayReflectionColor.r*reflectionMat.r +
			pointSpecularColor.r;
		outColor->g = 
			(pointColor.g*opacityMat.g + rayRefractionColor.g*(1 - opacityMat.g))*(1 - reflectionMat.g) + 
			rayReflectionColor.g*reflectionMat.g +
			pointSpecularColor.g;
		outColor->b = 
			(pointColor.b*opacityMat.b + rayRefractionColor.b*(1 - opacityMat.b))*(1 - reflectionMat.b) + 
			rayReflectionColor.b*reflectionMat.b +
			pointSpecularColor.b;
	}
	else //   
		*outColor = backgroundStatic;
}

void Scene::Test()
{
	//Rotate(&root.mTransform, 1.0f, 1.0f, 1.0f, PI/12);

	/*Quat q;
	MakeQuat(&q, 0.0f, 0.0f, 0.0f, PI/6);
	Matrix m, m2;
	MakeMatrix(&m, q);
	MultMatrices(&m2, root.mTransform, m);
	root.mTransform = m2;*/

	//RotateX(&root.mTransform, PI/6);

	//RotateZ(&root.mTransform, PI/12);

	/*Node *node = root.GetNode(0);
	RotateZ(&node->mTransform, PI/12);
	Scale(&node->mTransform, 1.05f, 1.05f, 0.95f);*/
}


void Scene::GetSceneInfo(SceneInfo *info)
{
	// prepare
	ReverseMatrix(&mViewReverse, mView);
	TriangleArray list;
	Matrix ident;
	MakeIdentity(&ident);
	root.Prepare(ident, true, &list);
	CreateCubes(&list);

	int cubesUsed = 0;
	int triTotal = 0;
	int triPeak = 0;

	for (int i = 0; i < cubeCount; i++)
		for (int j = 0; j < cubeCount; j++)
			for (int k = 0; k < cubeCount; k++)
			{
				PSimpleTriangleArray triList = cubes[i][j][k];
				if (triList)
				{
					cubesUsed++;
					triTotal += triList->GetCount();
					if (triList->GetCount() > triPeak)
						triPeak = triList->GetCount();
				}
			}

	info->trianglesCount = list.GetCount();
	info->lightsCount = lights.GetCount();
	info->camerasCount = cameras.GetCount();
	info->materialsCount = materials.GetCount();
	info->cubesPerLine = cubeCount;
	info->cubesTotal = cubeCount*cubeCount*cubeCount;
	info->cubesUsed = cubesUsed;
	info->triTotal = triTotal;
	info->triPeak = triPeak;
	info->triAverage = (REAL)triTotal/cubesUsed;

	// clear
	DeleteCubes();
}


void Scene::SetCurrentFrame(int frame)
{
	if (frame < 0 || frame >= totalFrame)
		throw L"Error SetCurrentFrame call";
	if (frame != currentFrame)
	{
		currentFrame = frame;
	}
}

void Scene::SetTotalFrame(int total)
{
	if (total <= 0)
		throw L"Error SetTotalFrame call";
	if (total != totalFrame)
	{
		totalFrame = total;
	}
}


void Scene::RenderTracingLine(const Vector &vScreenO, int j, Color *line)
{
	Vector vScreen, vdirView;
	for (int i = 0; i < nWidth; i++)
	{
		ScreenToWorldi(&vScreen, i, j);
		VectorSubtraction(&vdirView, vScreen, vScreenO);
		RayColor(vScreenO, vdirView, TRACING_LEVEL, &line[i]);
	}
}


void Scene::RenderTracingLinePlus(const Vector &vScreenO, int j, Color *line)
{
	Vector vScreen, vdirView;
	for (int i = -1; i <= nWidth; i++)
	{
		ScreenToWorldi(&vScreen, i, j);
		VectorSubtraction(&vdirView, vScreen, vScreenO);
		RayColor(vScreenO, vdirView, TRACING_LEVEL, &line[i + 1]);
	}
}


void Scene::RenderTracingLineAA(const Vector &vScreenO, int j, Color *line, Color *linePrev, Color *lineCurr, Color *lineNext)
{
	Vector vScreen, vdirView;
	for (int i = 0; i < nWidth; i++)
	{
		linePrev++;
		lineCurr++;
		lineNext++;

		REAL AmountDiff = 
			lineCurr->GetAmountDiff(*linePrev) + 
			lineCurr->GetAmountDiff(*lineNext) + 
			lineCurr->GetAmountDiff(*(lineCurr - 1)) + 
			lineCurr->GetAmountDiff(*(lineCurr + 1));

		if (AmountDiff < ADAPTIVE_ANTIALIASING_DIFF)
			*line = *lineCurr;
		else
		{
			const REAL d = (REAL)1/3;

			Color c, cTotal;
			cTotal = *lineCurr;

			//  
			ScreenToWorld(&vScreen, (REAL)i - d, (REAL)j - d);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);
			ScreenToWorld(&vScreen, (REAL)i, (REAL)j - d);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);
			ScreenToWorld(&vScreen, (REAL)i + d, (REAL)j - d);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);

			//   
			ScreenToWorld(&vScreen, (REAL)i - d, (REAL)j);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);
			ScreenToWorld(&vScreen, (REAL)i + d, (REAL)j);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);

			//  
			ScreenToWorld(&vScreen, (REAL)i - d, (REAL)j + d);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);
			ScreenToWorld(&vScreen, (REAL)i, (REAL)j + d);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);
			ScreenToWorld(&vScreen, (REAL)i + d, (REAL)j + d);
			VectorSubtraction(&vdirView, vScreen, vScreenO);
			RayColor(vScreenO, vdirView, TRACING_LEVEL, &c);
			cTotal.Add(c);

			line->r = cTotal.r/9;
			line->g = cTotal.g/9;
			line->b = cTotal.b/9;
		}

		line++;
	}
}


void Scene::RenderTracing()
{
	Callback(ssRenderStart, 0, 0, NULL);

	Callback(ssPrepareTriList, 0, 0, NULL);

	ReverseMatrix(&mViewReverse, mView);

	Vector oView, dirView, o(0, 0, 0);
	Transform(&oView, Vector(0, 0, 0), mViewReverse);

	TriangleArray list;
	Matrix ident;
	MakeIdentity(&ident);
	root.Prepare(ident, true, &list);

	Callback(ssCreateCubes, 0, 0, NULL);
	CreateCubes(&list);

	/*Color color;
	RayColor(Vector(0, 0, 0), Vector(-1, 0, -1), 0, &color);*/
	//RayColor(Vector(-2, -2, -2), Vector(0.5f, 1.1f, 0.7f), 0, &color);
	//RayColor(Vector(0, 0, 0), Vector(-1, -1, -1), 0, &color);
	//RayColor(oView, Vector (- oView.x, - oView.y, - oView.z), 0, &color);

#ifdef USE_ANTIALIASING

	Color *line, *linePrev, *lineCurr, *lineNext, *buf;
	line = new Color[nWidth];
	linePrev = new Color[nWidth + 2];
	lineCurr = new Color[nWidth + 2];
	lineNext = new Color[nWidth + 2];

	RenderTracingLinePlus(oView, -1, lineCurr);
	RenderTracingLinePlus(oView, 0, lineNext);

	for (int j = 0; j < nHeight; j++)
	{
		buf = linePrev;
		linePrev = lineCurr;
		lineCurr = lineNext;
		lineNext = buf;
		RenderTracingLinePlus(oView, j + 1, lineNext);

		RenderTracingLineAA(oView, j, line, linePrev, lineCurr, lineNext);

		for (int i = 0; i < nWidth; i++)
			//DrawPixel(i, j, line[i].GetRGB());
			buffer->SetColor(i, j, line[i]);
		
		bool cancel = false;
		Callback(ssRender, j + 1, nHeight, &cancel);
		if (cancel)
			break;
	}

	delete line;
	delete linePrev;
	delete lineCurr;
	delete lineNext;

#else

	Color *line = new Color[nWidth];
	for (int j = 0; j < nHeight; j++)
	{
		RenderTracingLine(oView, j, line);
		for (int i = 0; i < nWidth; i++)
			//DrawPixel(i, j, line[i].GetRGB());
			buffer->SetColor(i, j, line[i]);

		bool cancel = false;
		Callback(ssRender, j + 1, nHeight, &cancel);
		if (cancel)
			break;
	}
	delete line;

#endif

	//DrawText(L"Ray Tracer by pm01kps. For education purposes only.", buffer->GetWidth() - 360, buffer->GetHeight() - 20, RGB(0, 255, 0));

	DeleteCubes();
	
	Callback(ssRenderDone, 0, 0, NULL);
}
